import numpy as np
import random
import math


## Note this is RobustBandit algorithm in our paper
class restartrobust_reward:
    def __init__(self, attack_type, class_context, T):
        self.data = class_context
        self.T = T
        self.d = self.data.d
        self.attack = getattr(self.data, attack_type) 
        self.budget = self.data.attack_budget
        
    def restartrobust(self, delta=0.01, lamda=0.1):
        T = self.T
        d = self.data.d
        regret = np.zeros(T)
        
        
        # initialization for exp3 algo
        # the possible choices for C is in J
        K = len(self.data.fv[0])
        betaT = 0.01*math.sqrt( d*math.log((T*self.data.max_norm**2/lamda+1)/delta) ) + math.sqrt(lamda)
        gammaT = math.sqrt( 2*d*math.log(1+T/d/lamda) )
        H = betaT*gammaT / (math.sqrt(np.exp(1)-1)) * math.sqrt(T)
        J = [0] + [2**x for x in range(int(math.log(2*K*H)/math.log(2))+1)]
        Kexp = len(J)
        w = np.ones(Kexp)
        yhat = np.zeros(Kexp)
        gamma = min(1, math.sqrt( Kexp*math.log(Kexp) / ( (np.exp(1)-1) * T/H ) ) )
        p = gamma/ Kexp + (1-gamma) * w/sum(w)
        exp_r = 0
        t=0
        indexC = np.random.choice(Kexp)
        while t<T:
            if t % H == 0:
                gamma_explore = 0
                yhat[indexC] += gamma/ Kexp * exp_r / p[indexC]
                ymax = max(yhat)
                w = np.exp(yhat-ymax)
                gamma_explore = 0
                p = gamma/ Kexp + (1-gamma) * w/sum(w)
                indexC = np.random.choice(Kexp, p=p)
                c = J[indexC]
                xr = np.zeros(d)
                B = np.identity(d) * lamda
                B_inv = np.identity(d) / lamda
                theta_hat = np.zeros(d)
                exp_r = 0
            
            feature = self.data.fv[t]
            K = len(feature)
            
            # run exp3 to determine C^prime (c in the code)
            explore1 = 0.01*math.sqrt( d*math.log((t*self.data.max_norm**2/lamda+1)/delta) ) + math.sqrt(lamda)
            explore2 = math.sqrt(gamma_explore)
            ucb = np.zeros(K)
            width = np.zeros(K)
            for arm in range(K):
                # c is chosen by exp3
                ucb[arm] = feature[arm].dot(theta_hat)
                width[arm] = math.sqrt( feature[arm].T.dot(B_inv).dot(feature[arm]) )
            ucb_idx = ucb + width * (explore1 + explore2*c)
            pull = np.argmax(ucb_idx)
            gamma_explore += width[pull]**2
            
            if self.budget < 0:
                observe_r = self.data.random_reward(t,pull)
            else:
                observe_r, cost = self.attack(t, pull)
                self.budget -= cost
                
            # update linucb components
            B += np.outer(feature[pull], feature[pull])
            B_inv = np.linalg.inv(B)
            xr += feature[pull] * observe_r
            theta_hat = B_inv.dot(xr)
            regret[t] = regret[t-1] + self.data.optimal[t] - self.data.reward[t][pull]
            # update exp3 components
            exp_r += observe_r
            t += 1
        return regret